// GRAPH.CPP (MOUSE.EXE)
// (C) 1994 Ahpah Software Inc.
// Description: Interface to GENUS toolkits.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "defs.h"
#include "graph.h"
#include "system.h"

/* variables */
SIGNINT   gr_atype;      // memory allocation type
WORD      gr_mode;       // initialized graphics mode
GR_CURSOR gr_cursor;     // cursor information block
BOOLEAN   gr_alloc;      // allocation have occured
TEXT      gr_pal[GR_PALSIZE];  // palette for loading 256 color images


/* local functions */
STATIC BOOLEAN   gr_trymode( WORD, BOOLEAN );

/* gr_dogenus() reports an error on a toolkit call. (inline)
 */
INLINE VOID gr_dogenus( SIGNINT retval, STRING str )
    {
    if( retval != gxSUCCESS )
	gr_error( str, retval );
    }

/* gr_max() returns the maximum of two WORDS
 */
INLINE WORD gr_max( WORD a, WORD b )
    {
    if( a < b )
	return b;
    return a;
    }

/* gr_min() returns the maximum of two WORDS
 */
INLINE WORD gr_min( WORD a, WORD b )
    {
    if( a > b )
	return b;
    return a;
    }

/* gr_cursormove() will move the cursor to its new location
 */
VOID gr_cursormove( WORD x, WORD y )
    {
    WORD temploc[GR_VECT], newloc[GR_VECT] = { 0, 0, 0, 0 };
    WORD *curloc = gr_cursor.location;
    WORD xlen, ylen;
    GXHEADER temp;

    if( gr_cursor.type != GR_BITMAP )
	return;
    xlen = gr_cursor.fximage->xmax;
    ylen = gr_cursor.fximage->ymax;
    if( xlen / 2 < x )
	newloc[X1] = x - xlen / 2;
    if( ylen / 2 < y )
	newloc[Y1] = y - ylen / 2;
    newloc[X2] = newloc[X1] + xlen;
    newloc[Y2] = newloc[Y1] + ylen;
    temploc[X1] = gr_min( curloc[X1], newloc[X1] );
    temploc[Y1] = gr_min( curloc[Y1], newloc[Y1] );
    temploc[X2] = gr_max( curloc[X2], newloc[X2] );
    temploc[Y2] = gr_max( curloc[Y2], newloc[Y2] );
    if( gxCreateVirtual( gr_atype, &temp, gr_mode, XWID( temploc ), YWID( temploc ) ) != gxSUCCESS )
	gr_error( (STRING) "Could not create buffer." );
    CALL gxDisplayVirtual( RECT( temploc ), 0, &temp, 0, 0);
    CALL gxVirtualVirtual( &gr_cursor.savearea, 0, 0, xlen, ylen, &temp,
	curloc[X1] - temploc[X1], curloc[Y1] - temploc[Y1], gxSET );
    CALL gxVirtualVirtual( &temp, newloc[X1] - temploc[X1],
	newloc[Y1] - temploc[Y1], newloc[X1] - temploc[X1] + xlen,
	newloc[Y1] - temploc[Y1] + ylen, &gr_cursor.savearea, 0, 0, gxSET );
    CALL fxImageVirtual( &gr_cursor.fximage->fxheader, &temp,
	newloc[X1] - temploc[X1], newloc[Y1] - temploc[Y1] );
    CALL gxVirtualDisplay( &temp, 0, 0, RECT( temploc ), 0 );
    CALL gxDestroyVirtual( &temp );
    gr_vectorcopy( gr_cursor.location, newloc, GR_VECT );
    }

/* gr_cursorset() will set an internal cursor
 */
VOID gr_cursorset( SIGNINT style, LONG color )
    {
    gr_cursorstatus( GR_REMOVE );
    if( gr_cursor.type == GR_BITMAP )
	{
	gr_cursorstatus( GR_REMOVE );
	CALL grDisplayMouse( grSHOW );
	}
    gr_cursor.type = GR_INTERNAL;
    gr_cursor.style = style;
    gr_cursor.color = color;
    CALL grSetMouseStyle( style, color );
    gr_cursorstatus( GR_PLACE );
    }

/* gr_cursorset() will see a user bitmap cursor
 */
VOID gr_cursorset( GR_FXIMAGE *fxi )
    {
    if( gr_cursor.type == GR_BITMAP && fxi == gr_cursor.fximage )
	return;
    if( fxi->xmax > GR_CURSXMAX || fxi->ymax > GR_CURSYMAX )
	gr_error( (STRING) "Mouse cursor too large." );
    gr_cursorstatus( GR_REMOVE );
    if( gr_cursor.type == GR_INTERNAL )
	{
	CALL grDisplayMouse( grHIDE );
	gr_cursorstatus( GR_PLACE );
	}
    gr_cursor.type = GR_BITMAP;
    gr_cursor.fximage = fxi;
    gr_cursorstatus( GR_PLACE );
    }

/* gr_cursorstatus() will set the cursor on or off and update screen
 */
VOID gr_cursorstatus( WORD stat )
    {
    STATIC SIGNINT counter = 0;
    WORD *curloc = gr_cursor.location, xlen, ylen;
    SHORT x, y;

    switch( stat )
	{
	case GR_PLACE:
	    counter++;
	    if( gr_cursor.type != GR_BITMAP || counter != 1 )
		return;

	    CALL grGetMousePos( &x, &y );
	    xlen = gr_cursor.fximage->xmax;
	    ylen = gr_cursor.fximage->ymax;
	    curloc[X1] = 0;
	    curloc[Y1] = 0;
	    if( xlen / 2 < (WORD) x )
		curloc[X1] = (WORD) x - xlen / 2;
	    if( ylen / 2 < (WORD) y )
		curloc[Y1] = (WORD) y - ylen / 2;
	    curloc[X2] = curloc[X1] + xlen;
	    curloc[Y2] = curloc[Y1] + ylen;
	    CALL gxDisplayVirtual( RECT( curloc ), 0, &gr_cursor.savearea, 0, 0);
	    gr_fxshow( gr_cursor.fximage, UPLEFT( curloc ) );
	    break;
	case GR_REMOVE:
	    counter--;
	    if( gr_cursor.type != GR_BITMAP || counter != 0 )
		return;
	    CALL gxVirtualDisplay( &gr_cursor.savearea, 0, 0, RECT( gr_cursor.location ), 0 );
	    break;
	default:
	    break;
	}
    }

/* gr_ender() will close out video related functions
 */
VOID gr_ender()
    {
    if( gr_alloc )
	CALL gxDestroyVirtual( &gr_cursor.savearea );
    CALL gxClearDisplay( 0L, 0 );
    CALL gxSetMode( gxTEXT );
    CALL grStopMouse();
    CALL grSetEventMask( grENOEVENTS );
    CALL gxDone();
    }

/* gr_error() will display a fatal error
 */
VOID gr_error( STRING s )
    {
    gr_ender();
    sys_ender();
    printf( "Error: %s\n\n", s );
    exit( 1 );
    }

/* gr_error() will display a fatal error with GENUS exit code
 */
VOID gr_error( STRING s, SIGNINT n )
    {
    gr_ender();
    sys_ender();
    printf( "Error: %s (code %d)\n\n", s, n );
    exit( 1 );
    }

/* gr_fxcreate() will create an effects image from a GR_IMAGE
 */
VOID gr_fxcreate( GR_IMAGE *grimage, GR_FXIMAGE *fxi )
    {
    CALL fxSetKeyColor( 0L );
    if( fxCreateImage( &fxi->fxheader, &grimage->gxheader, gr_atype ) != gxSUCCESS )
	gr_error( (STRING) "Could not create image." );
    fxi->xmax = grimage->xmax;
    fxi->ymax = grimage->ymax;
    }

/* gr_fxfree() will free an effects image
 */
VOID gr_fxfree( GR_FXIMAGE *fxi, BOOLEAN freeit )
    {
    CALL fxDestroyImage( &fxi->fxheader );
    if( freeit )
	delete fxi;
    }

/* gr_fxshow() will show an effects image
 */
VOID gr_fxshow( GR_FXIMAGE *fxi, WORD x, WORD y )
    {
    CALL fxImageDisplay( &fxi->fxheader, x, y, 0 );
    }

/* gr_getevent() will wait for an event
 */
VOID gr_getevent( GREVENT *event )
    {
    CALL grSetEventMask( (SIGNINT) (grELPRESS|grERPRESS|grEMOUSEMOVE|grEKEYBOARD) );
    while( grGetEvent( event ) != gxSUCCESS )
	;
    CALL grSetEventMask( grENOEVENTS );
    gr_cursormove( (WORD) event->curx, (WORD) event->cury );
    }

/* gr_imagefree() will free a loaded image
 */
VOID gr_imagefree( GR_IMAGE *image, BOOLEAN freeit  )
    {
    CALL gxDestroyVirtual( &image->gxheader );
    if( freeit )
	delete image;
    }

/* gr_imageload() will load any supported image type
 */
VOID gr_imageload( GR_IMAGE *image, STRING fn )
    {
    IMGINHDR imghead;

    if( imgFileGetHeader( fn, 0, &imghead, gr_pal ) != gxSUCCESS )
	gr_error( (STRING) "Could not load image header." );
    image->xmax = imghead.Width;
    image->ymax = imghead.Height;
    if( gxCreateVirtual( gr_atype, &image->gxheader, gr_mode, image->xmax, image->ymax ) != gxSUCCESS )
	gr_error( (STRING) "Could not create buffer." );
    if( imgFileConvert( fn, 0, &image->gxheader ) != gxSUCCESS )
	gr_error( (STRING) "Could not load image data." );
    }

#if 0
/* gr_imagemake() will make a virtual buffer in GR_IMAGE form
 */
GR_IMAGE *gr_imagemake( WORD xlen, WORD ylen )
    {
    GR_IMAGE *image = new GR_IMAGE;

    if( gxCreateVirtual( gr_atype, &image->gxheader, gr_mode, xlen, ylen ) != gxSUCCESS )
	gr_error( (STRING) "Could not allocate buffer." );
    image->xmax = xlen;
    image->ymax = ylen;
    return image;
    }
#endif

/* gr_imageshow() will display an image
 */
VOID gr_imageshow( GR_IMAGE *image, WORD *vect )
    {
    GXHEADER newgx, *mhead;
    WORD xlen, ylen, x, y;

    xlen = vect[X2] - vect[X1] + 1;
    ylen = vect[Y2] - vect[Y1] + 1;
    x = vect[X1];
    y = vect[Y1];
    mhead = &image->gxheader;
    if( abs( xlen - image->xmax ) > 2 || abs( ylen - image->ymax ) > 2 )
	{
	if( gxCreateVirtual( gr_atype, &newgx, gr_mode, xlen, ylen ) != gxSUCCESS )
	    gr_error( (STRING) "Could not create buffer." );
	CALL gxVirtualScale( &image->gxheader, &newgx );
	mhead = &newgx;
	}
    else
	{
	xlen = image->xmax;
	ylen = image->ymax;
	}
    GR_CURSOROFF();
    CALL gxVirtualDisplay( mhead, 0, 0, x, y, x + xlen - 1, y + ylen - 1, 0 );
    GR_CURSORON();
    if( mhead == &newgx )
	CALL gxDestroyVirtual( &newgx );
    }

/* gr_mousebounds() will set the mouse boundaries
 */
VOID gr_mousebounds( WORD x1, WORD y1, WORD x2, WORD y2 )
    {
    CALL grSetMouseBounds( x1, y1, x2, y2 );
    }

/* gr_mousein() will check to see if the mouse is in a certain region
 */
BOOLEAN gr_mousein( GREVENT *event, WORD x1, WORD y1, WORD x2, WORD y2 )
    {
    if( x1 <= (WORD) event->curx && (WORD) event->curx <= x2 &&
	y1 <= (WORD) event->cury && (WORD) event->cury <= y2 )
	return YES;
    return NO;
    }

/* gr_starter() will initialize graphics routines.
 */
VOID gr_starter()
    {
    gr_dogenus( gxInit(), "Could not intialize GX Kernel." );
    gr_atype = gxCMM;
    if( !gr_trymode( gxVESA_112, NO ) )
	if( !gr_trymode( gxVESA_111, NO ) )
	    if( !gr_trymode( gxVESA_110, NO ) )
		CALL gr_trymode( gxVESA_101, YES );
    if( gr_mode <= gxVESA_103 )
	{
	CALL gxGetConvertPalette( gxPAL5, gr_pal );
	CALL gxSetDisplayPalette( gr_pal );
	CALL gxSetDitherMatrix( gxNODITHER );
	}
    if( grInitMouse() != gxSUCCESS )
	{
	CALL gxSetMode( gxTEXT );
	gr_error( (STRING) "No mouse found." );
	}
    CALL grSetMouseStyle( grCARROW, gxRGBPacked( 255, 255, 255 ) );
    CALL grDisplayMouse( grSHOW );
    CALL grTrackMouse( grTRACK );
    gr_mousebounds( 0, 0, GR_XMAX - 1, GR_YMAX - 1 );
    if( gxCreateVirtual( gr_atype, &gr_cursor.savearea, gr_mode, GR_CURSXMAX, GR_CURSYMAX ) != gxSUCCESS )
	gr_error( (STRING) "Could not create buffer." );
    gr_alloc = YES;
    }

/* gr_trymode() will attempt to initialize a video mode
 */
BOOLEAN gr_trymode( WORD mode, BOOLEAN doabort )
    {
    if( doabort )
	{
	gr_dogenus( gxSetDisplay( mode ), "Graphics mode not supported." );
	gr_dogenus( gxSetMode( gxGRAPHICS ), "Could not switch to graphics mode." );
	gr_mode = mode;
	return YES;
	}
    else
	{
	if( gxSetDisplay( mode ) == gxSUCCESS )
	    if( gxSetMode( gxGRAPHICS ) == gxSUCCESS )
		{
		gr_mode = mode;
		return YES;
		}
	}
    return NO;
    }


/* gr_vectorcopy() will copy one vector onto another
 */
VOID gr_vectorcopy( WORD *dest, WORD *src, WORD len )
    {
    memcpy( dest, src, len * sizeof( WORD ) );
    }

